// *.rtf -> *.tad converter

#include <vector>
#include "environment.h"
using namespace std;

extern void fusen_fontattr();
extern void fusen_fontsize();
extern void fusen_alignment();
extern void fusen_txtcolor();
extern void fusen_fonttype();
extern void fusen_decoration(unsigned char subid);
extern void fusen_soeji();
extern void fusen_tabindent();

extern bool tabChanged; // x֌WEɒ
extern map<int,Font> fnt;

vector<Environment> env;

Environment* Environment::_instance = 0;

Font::Font()
{
  charset = 128;
  font_type.uw = 0x00c0; 
}

// --------------------

Environment::Environment()
{
  int i;
  font_attr.uw = 0x8000; // normal
  font_size.uw = 0x8000 + 20*10; // 10pt
  nfont = 0;
  colorval.ul = 0x10000000; // black
  for (i=0;i<256;i++){
    is_deco[i] = false; // decorating now?
  }
  alignment = 0; // left-align
  is_soeji = false;
  ind_left.w = ind_first.w = ind_right.w = 0; // indentation
  unicode_ignore = 1; // num of ignored bytes after 'u' tag
}

Environment::Environment(const Environment& source)
{
  int i;
  font_attr.uw = source.font_attr.uw;
  font_size.uw = source.font_size.uw;
  nfont = source.nfont;
  colorval.ul = source.colorval.ul;
  for (i=0;i<256;i++){
    is_deco[i] = source.is_deco[i];
  }
  alignment = source.alignment;
  is_soeji = source.is_soeji;
  pat_soeji = source.pat_soeji;
  ind_left.w = source.ind_left.w;
  ind_first.w = source.ind_first.w;
  ind_right.w = source.ind_right.w;
  unicode_ignore = source.unicode_ignore;
}

Environment* Environment::current()
{
  if (_instance == 0){
    _instance = new Environment;
  }
  return _instance;
}

void Environment::outputCurrentEnvironment()
{
  Environment& newenv = *env.rbegin();
  static bool firstTime = true;
  
  // decoration
  for (int i=0;i<256;i++){
    if (current()->is_deco[i]){
      if (newenv.is_deco[i]){
        // true and true
        if (newenv.pat_deco[i] != current()->pat_deco[i]){
          current()->pat_deco[i] = newenv.pat_deco[i];
          fusen_decoration(i);
        }
      }
      else {
        // true(old) and false(new)
        current()->is_deco[i] = false;
        current()->pat_deco[i] = newenv.pat_deco[i];
        fusen_decoration(i+1);
      }
    }
    else {
      if (newenv.is_deco[i]){
        // false(old) and true(new)
        current()->is_deco[i] = true;
        current()->pat_deco[i] = newenv.pat_deco[i];
        fusen_decoration(i);
      }
      else {
        current()->pat_deco[i] = newenv.pat_deco[i];
      }
    }
  }
  // others
  if (newenv.font_attr.uw != current()->font_attr.uw){
    current()->font_attr.uw = newenv.font_attr.uw;
    fusen_fontattr();
  }
  if (newenv.font_size.uw != current()->font_size.uw || firstTime){
    current()->font_size.uw = newenv.font_size.uw;
    fusen_fontsize();
  }
  if (newenv.nfont != current()->nfont || firstTime){
    if (fnt[current()->nfont].font_type.uw != fnt[newenv.nfont].font_type.uw
        || firstTime){
      current()->nfont = newenv.nfont;
      fusen_fonttype();
    }
    else {
      current()->nfont = newenv.nfont;
    }
  }
  if (newenv.colorval.ul != current()->colorval.ul){
    current()->colorval.ul = newenv.colorval.ul;
    fusen_txtcolor();
  }
  if (newenv.alignment != current()->alignment){
    current()->alignment = newenv.alignment;
    fusen_alignment();
  }
  if (newenv.is_soeji || current()->is_soeji){
    current()->is_soeji = newenv.is_soeji;
    fusen_soeji();
  }
  if (newenv.ind_left.w != current()->ind_left.w
      || newenv.ind_right.w != current()->ind_right.w
      || newenv.ind_first.w != current()->ind_first.w){
    current()->ind_left.w = newenv.ind_left.w;
    current()->ind_right.w = newenv.ind_right.w;
    current()->ind_first.w = newenv.ind_first.w;
    fusen_tabindent();
    //tabChanged = false; // x֌WEɒ
  }
  
  firstTime = false;
}

