// *.rtf -> *.tad converter

using namespace std;

#include <iostream>
#include <sstream>
#include <string>
#include <map>
#include <vector>
#include <mbstring.h>
#include <io.h>
#include <fcntl.h>
#include "rtf2tad.h"
#include "environment.h"
#include "errmsg.h"
#include "rtfcoms.h"
#include "big5tocns.h"
#include "ibmmap.h"

typedef union {
  unsigned int mbch;
  struct { char hb; char lb; } byte;
} MBC;

map_elements me[] = {
  // ^O
  {"rtf",&nop},
  {"ansi",&nop},
  {"ansicpg",&nop},
  {"deff",&nop},
  {"deflang",&nop},
  {"deflangfe",&nop},
  {"viewkind",&nop},
  {"lang",&nop},
  // ꏏ̊̕܂ŊSɖ^O
  {"info",&tag_invalid},
  {"stylesheet",&tag_invalid},
  {"header",&tag_invalid},
  {"footer",&tag_invalid},
  // ^O
  {"par",&tag_par},
  {"line",&tag_line},
  {"page",&tag_page},
  {"rquote",&tag_rquote},
  {"lquote",&tag_rquote},
  {"rdblquote",&tag_rdblquote},
  {"ldblquote",&tag_ldblquote},
  {"emdash",&tag_emdash},
  {"endash",&tag_endash},
  {"b",&tag_bold},
  {"i",&tag_italic},
  {"fs",&tag_fontsize},
  {"tab",&tag_tab},
  {"qr",&tag_ralign},
  {"ql",&tag_lalign},
  {"qc",&tag_calign},
  {"pard",&tag_pard},
  {"colortbl",&tag_colortbl},
  {"cf",&tag_color},
  {"fonttbl",&tag_fonttbl},
  {"f",&tag_font},
  {"uc",&tag_uni_ignorebytes},
  {"u",&tag_unichar},
  // underlines (unified, but will be variated later...)
  {"ul",&tag_underline},
  {"uld",&tag_underline},
  {"uldash",&tag_underline},
  {"uldashd",&tag_underline},
  {"uldashdd",&tag_underline},
  {"uldb",&tag_underline},
  {"ulth",&tag_underline},
  {"ulwave",&tag_underline},
  // ---
  {"ulnone",&tag_no_underline},
  {"chshdng",&tag_shading},
  {"shad",&tag_shadow},
  {"outl",&tag_outline},
  {"super",&tag_super},
  {"sub",&tag_sub},
  {"nosupersub",&tag_soejioff},
  {"plain",&tag_plain},
  {"li",&tag_leftindent},
  {"fi",&tag_firstindent},
  {"ri",&tag_rightindent},
  {"pntext",&tag_pntext},
  {"listtext",&tag_pntext},
  {"",NULL} // ͏Ȃ悤
};

map<string,void (*)(string)> funcmap;

extern vector<Environment> env;
vector<unsigned long> col;
map<int,Font> fnt;

// ȉx֌W(ɒ)
string late;
bool islate = false;

// -------------------------------------------------------

void print_geta()
{
  cerr << "ϊłȂ܂Bo͂܂B" << endl;
  Environment::current()->outputCurrentEnvironment();
  changeCharset(1);
  cout.write("\x2e\x22",2);
}

void registermap(map<string,void (*)(string)>& fm)
{
  int i = 0;
  while (me[i].ptr != NULL){
    fm[me[i].key] = me[i].ptr;
    i++;
  }
}

char getbyte()
{
  char ch;
  if (cin.eof()){
    throw ERR_EOF;
  }
  if (islate && !late.empty()){
    // []x̏ꍇ
    ch = late[0];
    late.erase(0,1);
  }
  else {
    ch = cin.get();
  }
  return ch;
}

void putbyte(char ch)
{
  cout.put(ch);
}

char peekbyte()
{
  if (cin.eof()){
    throw ERR_EOF;
  }
  if (islate && !late.empty()){
    // []x̏ꍇ
    return late[0];
  }
  else {
    return cin.peek();
  }
}

void changeCharset(unsigned char x)
{
  static unsigned char currentCharset = 1;
  
  if (currentCharset != x){
    putbyte(x+0x20U);
    putbyte('\xfe');
    currentCharset = x;
  }
  return;
}

void on_japanese_literal(char ch)
{
  // {(Shift-JIS)
  MBC mbchar;
  static bool mb_mode = false;
  static char oldch;
  
  if (mb_mode){
    // }`oCg2
    mbchar.byte.hb = ch;
    mbchar.byte.lb = oldch;
    mbchar.mbch = _mbcjmstojis(mbchar.mbch);
    
    if (mbchar.mbch != 0){
      Environment::current()->outputCurrentEnvironment();
      changeCharset(1);
      putbyte(mbchar.byte.hb);
      putbyte(mbchar.byte.lb);
    }
    else if (oldch == '\xfa' && ((unsigned char)ch) >= 0x40U
             && ((unsigned char)ch) <= 0xfcU){
      int n = ((unsigned char)ch) - 0x40U;
      if (ibm2tc_tbl_fa[n] != 0){
        Environment::current()->outputCurrentEnvironment();
        changeCharset((ibm2tc_tbl_fa[n] >> 16) - 0x20U);
        putbyte(ibm2tc_tbl_fa[n] & 0x00ff);
        putbyte((ibm2tc_tbl_fa[n] & 0xff00) >> 8);
      }
      else goto not_found;
    }
    else if (oldch == '\xfb' && ((unsigned char)ch) >= 0x40U
             && ((unsigned char)ch) <= 0xfcU){
      int n = ((unsigned char)ch) - 0x40U;
      if (ibm2tc_tbl_fb[n] != 0){
        Environment::current()->outputCurrentEnvironment();
        changeCharset((ibm2tc_tbl_fb[n] >> 16) - 0x20U);
        putbyte(ibm2tc_tbl_fb[n] & 0x00ff);
        putbyte((ibm2tc_tbl_fb[n] & 0xff00) >> 8);
      }
      else goto not_found;
    }
    else if (oldch == '\xfc' && ((unsigned char)ch) >= 0x40U
             && ((unsigned char)ch) <= 0x4bU){
      int n = ((unsigned char)ch) - 0x40U;
      if (ibm2tc_tbl_fc[n] != 0){
        Environment::current()->outputCurrentEnvironment();
        changeCharset((ibm2tc_tbl_fc[n] >> 16) - 0x20U);
        putbyte(ibm2tc_tbl_fc[n] & 0x00ff);
        putbyte((ibm2tc_tbl_fc[n] & 0xff00) >> 8);
      }
      else goto not_found;
    }
    else {
    not_found:
      print_geta();
      mb_mode = false;
      return;
    }
    mb_mode = false;
  }
  else {
    if (_ismbblead(ch) || ((unsigned char)ch) >= 0x80U){
      // }`oCg1
      oldch = ch;
      mb_mode = true;
    }
    else {
      // VOoCg
      mbchar.mbch = _mbcjmstojis(_mbbtombc(ch));
      if (mbchar.mbch != 0){
        Environment::current()->outputCurrentEnvironment();
        changeCharset(1);
        putbyte(mbchar.byte.hb);
        putbyte(mbchar.byte.lb);
      }
      else {
        print_geta();
      }
    }
  }
}

void on_hantaiji_literal(char ch)
{
  // Big5(ɑ̎)
  unsigned char big5[2],cns[3];
  static bool mb_mode = false;
  static char oldch;
  
  if (mb_mode){
    // }`oCg2
    big5[1] = (unsigned)ch;
    big5[0] = (unsigned)oldch;
    Big5toCNS(cns,big5);
    if (cns[0] == 0x21){
      Environment::current()->outputCurrentEnvironment();
      changeCharset(6);
      putbyte(cns[2]);
      putbyte(cns[1]);
    }
    else if (cns[0] == 0x22){
      Environment::current()->outputCurrentEnvironment();
      changeCharset(6);
      putbyte(cns[2]);
      putbyte(cns[1]+0x5f);
    }
    mb_mode = false;
  }
  else {
    // }`oCg1
    if (((unsigned char)ch) >= 0xa1U && ((unsigned char)ch) <= 0xfeU){
      oldch = ch;
      mb_mode = true;
    }
    else {
      on_japanese_literal(ch);
    }
  }
}

void on_kantaiji_literal(char ch)
{
  // GB-EUC(ȑ̎)
  static bool mb_mode = false;
  static char oldch;
  
  if (mb_mode){
    // }`oCg2
    if (((unsigned char)ch) < 0xa1U || ((unsigned char)oldch) < 0xa1U){
      print_geta();
    }
    else {
      unsigned char ku = ((unsigned char)oldch) - 0xa0U;
      unsigned char ten = ((unsigned char)ch) - 0xa0U;
      unsigned int z = (ku - 1) * 94 + (ten - 1);
      if (oldch == '\xa9' && ch == '\x96'){
        // ́(1-21cf̔ۂőp) - gppxȂ̂ŌʑΉ
        Environment::current()->outputCurrentEnvironment();
        changeCharset(1);
        cout.write("\xcf\x21",2);
      }
      else {
        Environment::current()->outputCurrentEnvironment();
        changeCharset(1);
        putbyte((z % 126) + 0x80U);
        putbyte((z / 126) + 0x21U);
      }
    }
    mb_mode = false;
  }
  else {
    // }`oCg1
    if (((unsigned)ch) <=  0x7fU){
      on_japanese_literal(ch);
    }
    else {
      oldch = ch;
      mb_mode = true;
    }
  }
}

void on_ansi_literal(char ch)
{
  MBC mbc;
  
  if (((unsigned char)ch) <= 0x7fU){
    /*changeCharset(1);
    switch (ch){
    case 0x5c:
      mbc.mbch = 0x2140U;
      break;
    case 0x7e:
      mbc.mbch = 0x2232U;
      break;
    default:*/
    on_japanese_literal(ch);
    mbc.mbch = 0;
    // }
  }
  else if (((unsigned char)ch) <= 0x9fU){
    // ISO-8859-1`(MicrosoftgH)
    const static unsigned short chcd[] = {
      0x2921U, 0x0000U, 0x2124U, 0x253bU, 0x784dU, 0x0000U, 0x2277U, 0x2278U,
      0x2130U, 0x2273U, 0x2467U, 0x7868U, 0xa92dU, 0x0000U, 0xaa76U, 0x0000U,
      0x0000U, 0x2146U, 0x2147U, 0x2148U, 0x2149U, 0x2126U, 0x7841U, 0x7843U,
      0x2232U, 0xa26fU, 0x2468U, 0x7869U, 0xa94dU, 0x0000U, 0xab76U, 0xaa73U,
      0,
    };
    const static unsigned short chst[] = {
      1, 0, 1, 16, 16, 0, 1, 1,
      1, 1, 16, 16, 1, 0, 1, 0,
      0, 1, 1, 1, 1, 1, 16, 16,
      1, 1, 16, 16, 1, 0, 1, 1,
      0,
    };
    if (chst[((unsigned char)ch) - 0x80U] != 0){
      Environment::current()->outputCurrentEnvironment();
      changeCharset(chst[((unsigned char)ch) - 0x80U]);
    }
    mbc.mbch = chcd[((unsigned char)ch) - 0x80U];
  }
  else if (((unsigned char)ch) <= 0xbbU){
    Environment::current()->outputCurrentEnvironment();
    changeCharset(16);
    mbc.mbch = 0x2121U + ((unsigned char)ch) + 0xa2U;
  }
  else {
    Environment::current()->outputCurrentEnvironment();
    changeCharset(16);
    mbc.mbch = 0x2121U + ((unsigned char)ch) + 0xa2U * 2;
  }
  
  if (mbc.mbch != 0){
    putbyte(mbc.byte.hb);
    putbyte(mbc.byte.lb);
  }
}

void on_symbol_literal(char ch)
{
  // Symbol & WingDings ̏(Ă̂ŏ͂)
  switch (ch){
  case '\xb7':
  case '\x6c':
  case '\x9f':
    // 
    Environment::current()->outputCurrentEnvironment();
    changeCharset(1);
    cout.write("\x40\x23",2);
    break;
  case '\xd8':
    // ̐[݂ȃ}[N
    Environment::current()->outputCurrentEnvironment();
    changeCharset(16);
    cout.write("\x79\x8d",2);
    break;
  case '\x6f':
  case '\x70':
  case '\x71':
  case '\x72':
  case '\xa8':
    // `
    Environment::current()->outputCurrentEnvironment();
    changeCharset(16);
    cout.write("\x5f\x89",2);
    break;
  case '\x6e':
    // `
    Environment::current()->outputCurrentEnvironment();
    changeCharset(17);
    cout.write("\xb1\x27",2);
    break;
  case '\x75':
  case '\x77':
    // H`
    Environment::current()->outputCurrentEnvironment();
    changeCharset(16);
    cout.write("\x57\x8a",2);
    break;
  case '\xfc':
    // `FbN}[N
    Environment::current()->outputCurrentEnvironment();
    changeCharset(1);
    cout.write("\x7b\x27",2);
    break;
  case '\xb2':
    // 
    Environment::current()->outputCurrentEnvironment();
    changeCharset(16);
    cout.write("\x5c\x8c",2);
    break;
  default:
    print_geta();
    break;
  }
  return;
}

void on_korean_literal(char ch)
{
  // EUC-KR(؍)
  static bool mb_mode = false;
  static char oldch;
  
  if (mb_mode){
    // }`oCg2
    if (((unsigned char)ch) < 0xa1U || ((unsigned char)oldch) < 0xa1U){
      print_geta();
    }
    else {
      unsigned char ku = ((unsigned char)oldch) - 0xa0U;
      unsigned char ten = ((unsigned char)ch) - 0xa0U;
      unsigned int z = (ku - 1) * 94 + (ten - 1);
      Environment::current()->outputCurrentEnvironment();
      changeCharset(1);
      putbyte((z % 126) + 0x80U);
      putbyte((z / 126) + 183);
    }
    mb_mode = false;
  }
  else {
    // }`oCg1
    if (((unsigned)ch) <=  0x7fU){
      on_japanese_literal(ch);
    }
    else {
      oldch = ch;
      mb_mode = true;
    }
  }
}

void on_literal(char ch)
{
  // []pntext̒x
  if (!late.empty() && !islate){
    islate = true;
    while (!late.empty()){
      on_anychar(getbyte());
    }
    islate = false;
  }
  
  switch (fnt[(*env.rbegin()).nfont].charset){
  case 128:
    // {(Shift-JIS)
    on_japanese_literal(ch);
    break;
  case 0:
    // ISO-8859-1
    on_ansi_literal(ch);
    break;
  case 134:
    // EUC-GB2312(ȑ̎)
    on_kantaiji_literal(ch);
    break;
  case 136:
    // Big5(ɑ̎)
    on_hantaiji_literal(ch);
    break;
  case 129:
    // ؍(EUC-KR)
    on_korean_literal(ch);
    break;
  case 2:
    on_symbol_literal(ch);
    break;
  default:
    print_geta();
    //on_japanese_literal(ch);
    break;
  }
  return;
}

void get_command(string *com,string *param)
{
  char ch;
  com->clear();
  param->clear();
  
  while (isalpha(peekbyte())){
    ch = getbyte();
    *com += ch;
  }
  while (isdigit(peekbyte()) || peekbyte() == '-'){
    ch = getbyte();
    *param += ch;
  }
  if (peekbyte() == ' '){
    getbyte();
  }
}

void on_command()
{
  string com, param;
  map<string,void (*)(string)>::iterator keyitr;
  char ch;
  int i,intch;
  
  get_command(&com,&param);
  // RTF^O̎s
  if (!com.empty()){
    keyitr = funcmap.find(com);
    if (keyitr != funcmap.end()){
      (*(keyitr->second))(param);
    }
    else {
      err_tagnotfound(com);
      funcmap[com] = &nop;
    }
  }
  else {
    ch = getbyte();
    switch (ch){
    case '\\':
    case '{':
    case '}':
      on_literal(ch);
      break;
    case '\'':
      {
        com.erase();
        for (i=0;i<2;i++){
          com += getbyte();
        }
        istringstream stin(com);
        stin >> hex >> intch;
        on_literal(intch);
      }
      break;
    case '*':
      // ł鑮BB
      voidtag();
      break;
    }
  }
}

/*void delenv()
{
  Environment oldenv,newenv;
  newenv = *env.rbegin();
  env.pop_back();
  oldenv = *env.rbegin();
  // decoration
  for (int i=0;i<256;i++){
    if (oldenv.is_deco[i]){
      if (newenv.is_deco[i]){
        // true and true
        if (newenv.pat_deco[i] != oldenv.pat_deco[i]){
          fusen_decoration(i);
        }
      }
      else {
        // true(old) and false(new)
        fusen_decoration(i);
      }
    }
    else {
      if (newenv.is_deco[i]){
        // false(old) and true(new)
        fusen_decoration(i+1);
      }
    }
  }
  // others
  if (newenv.font_attr.uw != oldenv.font_attr.uw){
    fusen_fontattr();
  }
  if (newenv.font_size.uw != oldenv.font_size.uw){
    fusen_fontsize();
  }
  if (newenv.nfont != oldenv.nfont){
    fusen_fonttype();
  }
  if (newenv.colorval.ul != oldenv.colorval.ul){
    fusen_txtcolor();
  }
  if (newenv.alignment != oldenv.alignment){
    fusen_alignment();
  }
  if (newenv.is_soeji || oldenv.is_soeji){
    fusen_soeji();
  }
  if (newenv.ind_left.w != oldenv.ind_left.w
      || newenv.ind_right.w != oldenv.ind_right.w
      || newenv.ind_first.w != oldenv.ind_first.w){
    fusen_tabindent();
    tabChanged = false;
  }
}*/

void on_anychar(char ch)
{
  switch (ch){
  case '\\':
    // command
    on_command();
    break;
  case '{':
    // new environment
    env.push_back(*env.rbegin());
    break;
  case '}':
    // delete environment
    env.pop_back();
    break;
  case '\x0d':
  case '\x0a':
    break;
  case '\t':
    tag_tab("");
    break;
  default:
    // japanese or non-japanese?
    on_literal(ch);
    
    break;
  }
}

int main(int argc,char *argv[])
{
  registermap(funcmap);
  
  _setmode( _fileno( stdout ), _O_BINARY );
  
  Environment stdenv;
  env.push_back(stdenv);
  
  try {
    char ch;
    header();
    
    while (!cin.eof()){
      ch = getbyte();
      if (ch != '\x0' && ch != '\xff'){
        on_anychar(ch);
      }
    }
    
    footer();
  }
  catch (int intval) {
    err_parsing();
    return 1;
  }
  
  cerr << "ϊ܂B" << endl;
  
  return 0;
}
